"""PyTorch Binary Classification with Tabular data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..utils.pytorch import fit
from ..orm import Dataset
# External modules
import torch.nn as nn
from torch.optim import Adamax
import torchmetrics as tm
from sklearn.preprocessing import PowerTransformer, LabelBinarizer


def fn_build(features_shape, label_shape, **hp):
	model = nn.Sequential(
		nn.Linear(features_shape[-1], 12),
		nn.BatchNorm1d(12,12),
		nn.ReLU(),
		nn.Dropout(p=0.5),

		nn.Linear(12, label_shape[-1]),
		nn.Sigmoid()
	)
	return model


def fn_optimize(model, **hp):
	optimizer = Adamax(
		model.parameters(), lr=hp['learning_rate']
	)
	return optimizer


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	return fit(
		model, loser, optimizer, 
		train_features, train_label,
		eval_features, eval_label,
		epochs=hp['epoch_count'], batch_size=5,
		metrics=[tm.Accuracy(),tm.F1Score()]
	)


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=2):
	hyperparameters = {
		"learning_rate": [0.01]
		, "epoch_count": [10]
	}
	
	file_path = datum.get_path('sonar.csv')
	
	shared_dataset = Dataset.Tabular.from_path(
		file_path = file_path
	)

	pipeline = Pipeline(
		Input(
			dataset  = shared_dataset,
			encoders = Input.Encoder(
				PowerTransformer(method='yeo-johnson', copy=False),
				dtypes = ['float64']
			)
		),

		Target(
			dataset = shared_dataset,
			column  = 'object',
			encoder = Target.Encoder(LabelBinarizer(sparse_output=False))
		),

		Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)
	)
	
	experiment = Experiment(
		Architecture(
			library           = "pytorch"
			, analysis_type   = "classification_binary"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment


from pprint import pformat
from textwrap import dedent
from operator import ge, le
import tensorflow as tf
from math import ceil
import numpy as np

## --- Default Architectures ---
# `**hp` must be included

# `select_fn_lose()`
def lose_regression(**hp):
    loser = tf.keras.losses.MeanAbsoluteError()
    return loser

def lose_binary(**hp):
    loser = tf.keras.losses.BinaryCrossentropy()
    return loser

def lose_multiclass(**hp):
    loser = tf.keras.losses.CategoricalCrossentropy()
    return loser

# `select_fn_optimize()`
def optimize(**hp):
    optimizer = tf.keras.optimizers.Adamax(learning_rate=0.01)
    return optimizer

# `select_fn_predict()`
def predict_multiclass(model, features):
    # Shows the probabilities of each class coming out of softmax neurons:
    # array([[9.9990356e-01, 9.6374511e-05, 3.3754202e-10],...])
    probabilities = model.predict(features)
    # This is the official keras replacement for multiclass `.predict_classes()`
    # Returns one ordinal array per sample: `[[0][1][2][3]]` 
    prediction = np.argmax(probabilities, axis=-1)
    return prediction, probabilities

def predict_binary(model, features):
    # Sigmoid output is between 0 and 1.
    # It's not technically a probability, but it is still easy to interpret.
    probability = model.predict(features)
    # This is the official keras replacement for binary classes `.predict_classes()`.
    # Returns one array per sample: `[[0][1][0][1]]`.
    prediction = (probability > 0.5).astype("int32")
    return prediction, probability

def predict_regression(model, features):
    prediction = model.predict(features)
    # ^ Output is a single value, not `probability, prediction`
    return prediction


class TrainingCallback():
    class MetricCutoff(tf.keras.callbacks.Callback):
        """
        - Worried that these inner functions are not pickling during multi-processing.
          stackoverflow.com/a/8805244/5739514
        """
        def __init__(self, thresholds:list):
            """
            # Tested with keras:2.4.3, tensorflow:2.3.1
            # `thresholds` is list of dictionaries with 1 dict per metric.
            metrics_cuttoffs = [
                {"metric":"val_acc", "cutoff":0.94, "above_or_below":"above"},
                {"metric":"acc", "cutoff":0.90, "above_or_below":"above"},
                {"metric":"val_loss", "cutoff":0.26, "above_or_below":"below"},
                {"metric":"loss", "cutoff":0.30, "above_or_below":"below"},
            ]
            # Only stops training early if all user-specified metrics are satisfied.
            # `above_or_below`: where 'above' means `>=` and 'below' means `<=`.
            """
            self.thresholds = thresholds
            

        def on_epoch_end(self, epoch, logs=None):
            logs = logs or {}
            # Check each user-defined threshold to see if it is satisfied.
            for threshold in self.thresholds:
                metric = logs.get(threshold['metric'])
                if (metric is None):
                    raise Exception(dedent(f"""
                    Yikes - The metric named '{threshold['metric']}' not found when running `logs.get('{threshold['metric']}')`
                    during `TrainingCallback.MetricCutoff.on_epoch_end`.
                    """))
                cutoff = threshold['cutoff']

                above_or_below = threshold['above_or_below']
                if (above_or_below == 'above'):
                    statement = ge(metric, cutoff)
                elif (above_or_below == 'below'):
                    statement = le(metric, cutoff)
                else:
                    raise Exception(dedent(f"""
                    Yikes - Value for key 'above_or_below' must be either string 'above' or 'below'.
                    You provided:{above_or_below}
                    """))

                if (statement == False):
                    break # Out of for loop.
                    
            if (statement == False):
                pass # Thresholds not satisfied, so move on to the next epoch.
            elif (statement == True):
                # However, if the for loop actually finishes, then all metrics are satisfied.
                print(
                    f":: Epoch #{epoch} ::\n" \
                    f"Congratulations - satisfied early stopping thresholds defined in `MetricCutoff` callback:\n"\
                    f"{pformat(self.thresholds)}\n"
                )
                self.model.stop_training = True


## --- Batching ---
# Used for a manual training loop without Keras.
def batcher(features:object, labels:object, batch_size:int=5):
    """
    - `np.array_split` allows for subarrays to be of different sizes, which is rare.
      https://numpy.org/doc/stable/reference/generated/numpy.array_split.html 
    - If there is a remainder, it will evenly distribute samples into the other arrays.
    - Have not tested this with >= 3D data yet.
    """
    rows_per_batch = ceil(features.shape[0]/batch_size)

    batched_features = np.array_split(features, rows_per_batch)
    batched_features = np.array(batched_features, dtype=object)

    batched_labels = np.array_split(labels, rows_per_batch)
    batched_labels = np.array(batched_labels, dtype=object)
    return batched_features, batched_labels



"""Sets defaults when assembling architecture in Queue object"""
from . import tensorflow, pytorch
from textwrap import dedent


def select_fn_lose(
    library:str,
    analysis_type:str
):      
    fn_lose = None
    if (library == 'keras'):
        if (analysis_type == 'regression'):
            fn_lose = tensorflow.lose_regression
        elif (analysis_type == 'classification_binary'):
            fn_lose = tensorflow.lose_binary
        elif (analysis_type == 'classification_multi'):
            fn_lose = tensorflow.lose_multiclass
    elif (library == 'pytorch'):
        if (analysis_type == 'regression'):
            fn_lose = pytorch.lose_regression
        elif (analysis_type == 'classification_binary'):
            fn_lose = pytorch.lose_binary
        elif (analysis_type == 'classification_multi'):
            fn_lose = pytorch.lose_multiclass
    # After each of the predefined approaches above, check if it is still undefined.
    if fn_lose is None:
        raise Exception(dedent("""
        Yikes - You did not provide a `fn_lose`,
        and we don't have an automated function for your combination of 'library' and 'analysis_type'
        """))
    return fn_lose


def select_fn_optimize(library:str):
    fn_optimize = None
    if (library == 'keras'):
        fn_optimize = tensorflow.optimize
    elif (library == 'pytorch'):
        fn_optimize = pytorch.optimize
    # After each of the predefined approaches above, check if it is still undefined.
    if (fn_optimize is None):
        raise Exception(dedent("""
        Yikes - You did not provide a `fn_optimize`,
        and we don't have an automated function for your 'library'
        """))
    return fn_optimize


def select_fn_predict(
    library:str,
    analysis_type:str
):
    fn_predict = None
    if (library == 'keras'):
        if (analysis_type == 'classification_multi'):
            fn_predict = tensorflow.predict_multiclass
        elif (analysis_type == 'classification_binary'):
            fn_predict = tensorflow.predict_binary
        elif (analysis_type == 'regression'):
            fn_predict = tensorflow.predict_regression
    elif (library == 'pytorch'):
        if (analysis_type == 'classification_multi'):
            fn_predict = pytorch.predict_multiclass
        elif (analysis_type == 'classification_binary'):
            fn_predict = pytorch.predict_binary
        elif (analysis_type == 'regression'):
            fn_predict = pytorch.predict_regression

    # After each of the predefined approaches above, check if it is still undefined.
    if fn_predict is None:
        raise Exception(dedent("""
        Yikes - You did not provide a `fn_predict`,
        and we don't have an automated function for your combination of 'library' and 'analysis_type'
        """))
    return fn_predict



"""TensorFlow Binary Classification with Sequence data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..orm import Dataset
from ..utils.config import app_folders, create_folder
# External modules
from os import path
from numpy import save as np_save
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import StandardScaler
import numpy as np


def fn_build(features_shape, label_shape, **hp):    
	# features_shape = samples, rows, columns
	# label_shape = samples, columns
	m = tf.keras.models.Sequential()
	m.add(l.Input(shape=features_shape))
	m.add(l.LSTM(hp['neuron_count']))
	m.add(l.Dense(units=label_shape[-1], activation='sigmoid'))
	return m


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	model.compile(
		loss        = loser
		, optimizer = optimizer
		, metrics   = ['accuracy']
	)
	model.fit(
		train_features, train_label
		, validation_data = (eval_features, eval_label)
		, verbose         = 0
		, batch_size      = hp['batch_size']
		, epochs          = hp['epochs']
		, callbacks       = [tf.keras.callbacks.History()]
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=3):
	df = datum.to_df('epilepsy.parquet')
	# testing FeatureInterpolater 3D.
	df['sensor_1'][10]   = np.NaN
	df['sensor_1'][0]    = np.NaN
	df['sensor_150'][80] = np.NaN
	df['sensor_152'][22] = np.NaN
	df['sensor_170'][0]  = np.NaN
	
	label_df = df[['seizure']]
	label_dataset = Dataset.Tabular.from_df(label_df)

	sensor_arr3D = df.drop(columns=['seizure']).to_numpy().reshape(1000,178,1).astype('float64')
	# Just testing all ingestion scenarios
	if (fold_count is None):
		feature_dataset = Dataset.Sequence.from_numpy(sensor_arr3D)
	else:
		path_models_cache = app_folders['cache_tests']
		create_folder(path_models_cache)
		path_file = f"temp_arr.npy"
		path_full = path.join(path_models_cache,path_file)
		np_save(path_full, sensor_arr3D, allow_pickle=True)
		feature_dataset = Dataset.Sequence.from_numpy(
			arr3D_or_npyPath=path_full, ingest=False, retype='float64'
		)

	hyperparameters = dict(neuron_count= [18], batch_size=[8], epochs=[5])

	pipeline = Pipeline(
		Input(
			dataset       = feature_dataset,
			interpolaters = Input.Interpolater(dtypes="float64"),
			encoders      = Input.Encoder(sklearn_preprocess=StandardScaler())
		),
		
		Target(
			dataset = label_dataset
		),
		
		Stratifier(
			size_test       = 0.12, 
			size_validation = 0.22,
			fold_count      = fold_count
		)    
	)
	experiment = Experiment(
		Architecture(
			library           = "keras"
			, analysis_type   = "classification_binary"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
		)
	)
	return experiment



"""TensorFlow Binary Classification with Tabular data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..orm import Dataset
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import LabelBinarizer, PowerTransformer


def fn_build(features_shape, label_shape, **hp):
	model = tf.keras.models.Sequential()
	model.add(l.Input(shape=features_shape))
	model.add(l.Dense(15, activation='relu', kernel_initializer='he_uniform'))
	model.add(l.Dropout(0.30))
	model.add(l.Dense(units=label_shape[-1], activation='sigmoid', kernel_initializer='glorot_uniform'))
	return model


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	model.compile(
		loss = loser
		, optimizer = optimizer
		, metrics = ['accuracy']
	)
	model.fit(
		train_features, train_label
		, validation_data = (eval_features, eval_label)
		, verbose = 0
		, batch_size = 3
		, epochs = 15
		, callbacks = [tf.keras.callbacks.History()]
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=2):
	## test for `None`
	hyperparameters = {"neuron_count":[15], "epochs":[15]}
	# hyperparameters = None

	file_path = datum.get_path('sonar.csv')

	# testing overlap of hashes
	if (fold_count is not None):
		name = 'rocks n radio'
	else:
		name = 'mines'

	shared_dataset = Dataset.Tabular.from_path(
		file_path = file_path
		, name = name
	)
	
	pipeline = Pipeline(
		Input(
			dataset  = shared_dataset,
			encoders = Input.Encoder(sklearn_preprocess=PowerTransformer())
		),
		
		Target(
			dataset = shared_dataset,
			column = 'object',
			encoder = Target.Encoder(sklearn_preprocess=LabelBinarizer())
		),
		
		Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)    
	)
	experiment = Experiment(
		Architecture(
			library           = "keras"
			, analysis_type   = "classification_binary"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment



"""TensorFlow Forecasting with Tabular data"""
# Internal modules
from ..mlops import Pipeline, Input, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..orm import Dataset
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import StandardScaler, RobustScaler
import numpy as np


def fn_build(features_shape, label_shape, **hp):
	m = tf.keras.models.Sequential()
	# `l.Input(shape=features_shape)` not working as expected.
	# Warns about multi-tensor layers.
	m(l.GRU(hp['neuron_count'], input_shape=features_shape, return_sequences=False, activation='tanh'))
	# Automatically flattens.
	m(l.Dense(label_shape[0]*label_shape[1]*hp['dense_multiplier'], activation='tanh'))
	m(l.Dropout(0.3))
	m(l.Dense(label_shape[0]*label_shape[1], activation='tanh'))
	m(l.Dropout(0.3))
	# Reshape to be 3D.
	m(l.Reshape(label_shape))
	
	return m


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	model.compile(
		loss=loser
		, optimizer=optimizer
		, metrics=['mean_squared_error']
	)
		
	model.fit(
		train_features, train_label
		, validation_data = (eval_features, eval_label,)
		, verbose = 0
		, batch_size = hp['batch_size']
		, epochs = hp['epochs']
		, callbacks = [tf.keras.callbacks.History()]
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=2):
	hyperparameters = {
		"neuron_count": [8]
		, "batch_size": [8]
		, "epochs": [12]
		, "dense_multiplier": [1]
	}
	
	df = datum.to_df('delhi_climate.parquet')
	df['temperature'][0] = np.NaN
	df['temperature'][13] = np.NaN
	
	dataset = Dataset.Tabular.from_df(dataframe=df)

	pipeline = Pipeline(
		inputs = Input(
			dataset  = dataset,
			interpolaters = Input.Interpolater(dtypes=['float64']),
			window = Input.Window(size_window=28, size_shift=14),
			encoders = [
				Input.Encoder(
					RobustScaler(),
					columns = ['wind', 'pressure']
				),
				Input.Encoder(
					StandardScaler(),
					dtypes = ['float64', 'int64']
				),
			]
		),
		
		stratifier = Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)    
	)

	experiment = Experiment(
		Architecture(
			library           = "keras"
			, analysis_type   = "regression"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment



from .wrangle import if_1d_make_2d
from .ingest import colIndices_from_colNames
from inspect import isclass
from warnings import catch_warnings
from scipy import sparse
from textwrap import dedent
import numpy as np


# Used during encoding validation.
categorical_encoders = [
    'OneHotEncoder', 'LabelEncoder', 'OrdinalEncoder', 
    'Binarizer', 'LabelBinarizer', 'MultiLabelBinarizer'
]

# Used by `sklearn.preprocessing.FunctionTransformer` to normalize images.
def div255(x): return x/255
def mult255(x): return x*255


def check_sklearn_attributes(sklearn_preprocess:object, is_label:bool):
    #This function is used by Featurecoder too, so don't put label-specific things in here.
    if (isclass(sklearn_preprocess)):
        raise Exception(dedent("""
            Yikes - The encoder you provided is a class name, but it should be a class instance.\n
            Class (incorrect): `OrdinalEncoder`
            Instance (correct): `OrdinalEncoder()`
        """))

    # Encoder parent modules vary: `sklearn.preprocessing._data` vs `sklearn.preprocessing._label`
    # Feels cleaner than this: stackoverflow.com/questions/14570802
    coder_type = str(type(sklearn_preprocess))
    if ('sklearn.preprocessing' not in coder_type):
        raise Exception(dedent("""
            Yikes - At this point in time, only `sklearn.preprocessing` encoders are supported.
            https://scikit-learn.org/stable/modules/classes.html#module-sklearn.preprocessing
        """))
    elif ('sklearn.preprocessing' in coder_type):
        if (not hasattr(sklearn_preprocess, 'fit')):    
            raise Exception(dedent("""
                Yikes - The `sklearn.preprocessing` method you provided does not have a `fit` method.\n
                Please use one of the uppercase methods instead.
                For example: use `RobustScaler` instead of `robust_scale`.
            """))

        if (hasattr(sklearn_preprocess, 'sparse')):
            if (sklearn_preprocess.sparse == True):
                try:
                    sklearn_preprocess.sparse = False
                    print(dedent("""
                        └── Info - System overriding user input to set `sklearn_preprocess.sparse=False`.
                            This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.
                    """))
                except:
                    raise Exception(dedent(f"""
                        Yikes - Detected `sparse==True` attribute of {sklearn_preprocess}.
                        System attempted to override this to False, but failed.
                        FYI `sparse` is True by default if left blank.
                        This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.\n
                        Please try again with False. For example, `OneHotEncoder(sparse=False)`.
                    """))

        if (hasattr(sklearn_preprocess, 'drop')):
            if (sklearn_preprocess.drop is not None):
                try:
                    sklearn_preprocess.drop = None
                    print(dedent("""
                        └── Info - System overriding user input to set `sklearn_preprocess.drop`.
                            System cannot handle `drop` yet when dynamically inverse_transforming predictions.
                    """))
                except:
                    raise Exception(dedent(f"""
                        Yikes - Detected `drop is not None` attribute of {sklearn_preprocess}.
                        System attempted to override this to None, but failed.
                    """))

        if (hasattr(sklearn_preprocess, 'copy')):
            if (sklearn_preprocess.copy == True):
                try:
                    sklearn_preprocess.copy = False
                    print(dedent("""
                        └── Info - System overriding user input to set `sklearn_preprocess.copy=False`.
                        This saves memory when concatenating the output of many encoders.
                    """))
                except:
                    raise Exception(dedent(f"""
                        Yikes - Detected `copy==True` attribute of {sklearn_preprocess}.
                        System attempted to override this to False, but failed.
                        FYI `copy` is True by default if left blank, which consumes memory.\n
                        Please try again with 'copy=False'.
                        For example, `StandardScaler(copy=False)`.
                    """))
        
        if (hasattr(sklearn_preprocess, 'sparse_output')):
            if (sklearn_preprocess.sparse_output == True):
                try:
                    sklearn_preprocess.sparse_output = False
                    print(dedent("""
                        └── Info - System overriding user input to set `sklearn_preprocess.sparse_output=False`.
                            This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.
                    """))
                except:
                    raise Exception(dedent(f"""
                        Yikes - Detected `sparse_output==True` attribute of {sklearn_preprocess}.
                        System attempted to override this to True, but failed.
                        Please try again with 'sparse_output=False'.
                        This would have generated 'scipy.sparse.csr.csr_matrix', causing Keras training to fail.\n
                        For example, `LabelBinarizer(sparse_output=False)`.
                    """))

        if (hasattr(sklearn_preprocess, 'order')):
            if (sklearn_preprocess.order == 'F'):
                try:
                    sklearn_preprocess.order = 'C'
                    print(dedent("""
                        └── Info - System overriding user input to set `sklearn_preprocess.order='C'`.
                            This changes the output shape of the 
                    """))
                except:
                    raise Exception(dedent(f"""
                        System attempted to override this to 'C', but failed.
                        Yikes - Detected `order=='F'` attribute of {sklearn_preprocess}.
                        Please try again with 'order='C'.
                        For example, `PolynomialFeatures(order='C')`.
                    """))

        if (hasattr(sklearn_preprocess, 'encode')):
            if (sklearn_preprocess.encode == 'onehot'):
                # Multiple options here, so don't override user input.
                raise Exception(dedent(f"""
                    Yikes - Detected `encode=='onehot'` attribute of {sklearn_preprocess}.
                    FYI `encode` is 'onehot' by default if left blank and it predictors in 'scipy.sparse.csr.csr_matrix',
                    which causes Keras training to fail.\n
                    Please try again with 'onehot-dense' or 'ordinal'.
                    For example, `KBinsDiscretizer(encode='onehot-dense')`.
                """))

        if (
            (is_label==True)
            and
            (not hasattr(sklearn_preprocess, 'inverse_transform'))
        ):
            print(dedent("""
                Warning - The following encoders do not have an `inverse_transform` method.
                It is inadvisable to use them to encode Labels during training, 
                because you may not be able to programmatically decode your raw predictions 
                when it comes time for inference (aka non-training predictions):

                [Binarizer, KernelCenterer, Normalizer, PolynomialFeatures, SplineTransformer]
            """))

        """
        - Binners like 'KBinsDiscretizer' and 'QuantileTransformer'
            will place unseen observations outside bounds into existing min/max bin.
        - I assume that someone won't use a custom FunctionTransformer, for categories
            when all of these categories are available.
        - LabelBinarizer is not threshold-based, it's more like an OHE.
        """
        only_fit_train = True
        stringified_coder = str(sklearn_preprocess)
        is_categorical = False
        for c in categorical_encoders:
            if (stringified_coder.startswith(c)):
                only_fit_train = False
                is_categorical = True
                break

        return sklearn_preprocess, only_fit_train, is_categorical


def fit_dynamicDimensions(sklearn_preprocess:object, samples_to_fit:object):
    """
    - There are 17 uppercase sklearn encoders, and 10 different data types across float, str, int 
      when consider negatives, 2D multiple columns, 2D single columns.
    - Different encoders work with different data types and dimensionality.
    - This function normalizes that process by coercing the dimensionality that the encoder wants,
      and erroring if the wrong data type is used. The goal in doing so is to return 
      that dimensionality for future use.

    - `samples_to_transform` is pre-filtered for the appropriate `matching_columns`.
    - The rub lies in that if you have many columns, but the encoder only fits 1 column at a time, 
      then you return many fits for a single type of preprocess.
    - Remember this is for a single Featurecoder that is potential returning multiple fits.

    - UPDATE: after disabling LabelBinarizer and LabelEncoder from running on multiple columns,
      everything seems to be fitting as "2D_multiColumn", but let's keep the logic for new sklearn methods.
    """
    fitted_encoders   = []
    incompatibilities = {
        "string": [
            "KBinsDiscretizer", "KernelCenterer", "MaxAbsScaler", 
            "MinMaxScaler", "PowerTransformer", "QuantileTransformer", 
            "RobustScaler", "StandardScaler"
        ]
        , "float": ["LabelBinarizer"]
        , "numeric array without dimensions both odd and square (e.g. 3x3, 5x5)": ["KernelCenterer"]
    }

    with catch_warnings(record=True) as w:
        try:
            #`samples_to_fit` is coming in as 2D.
            # Remember, we are assembling `fitted_encoders` dict, not accessing it.
            fit_encoder = sklearn_preprocess.fit(samples_to_fit)
            fitted_encoders.append(fit_encoder)
        except:
            # At this point, "2D" failed. It had 1 or more columns.
            try:
                width = samples_to_fit.shape[1]
                if (width > 1):
                    # Reshape "2D many columns" to “3D of 2D single columns.”
                    samples_to_fit = samples_to_fit[None].T                    
                    # "2D single column" already failed. Need it to fail again to trigger except.
                elif (width == 1):
                    # Reshape "2D single columns" to “3D of 2D single columns.”
                    samples_to_fit = samples_to_fit.reshape(1, samples_to_fit.shape[0], 1)    
                # Fit against each 2D array within the 3D array.
                for i, arr in enumerate(samples_to_fit):
                    fit_encoder = sklearn_preprocess.fit(arr)
                    fitted_encoders.append(fit_encoder)
            except:
                # At this point, "2D single column" has failed.
                try:
                    # So reshape the "3D of 2D_singleColumn" into "2D of 1D for each column."
                    # This transformation is tested for both (width==1) as well as (width>1). 
                    samples_to_fit = samples_to_fit.transpose(2,0,1)[0]
                    # Fit against each column in 2D array.
                    for i, arr in enumerate(samples_to_fit):
                        fit_encoder = sklearn_preprocess.fit(arr)
                        fitted_encoders.append(fit_encoder)
                except:
                    raise Exception(dedent(f"""
                        Yikes - Encoder failed to fit the columns you filtered.\n
                        Either the data is dirty (e.g. contains NaNs),
                        or the encoder might not accept negative values (e.g. PowerTransformer.method='box-cox'),
                        or you used one of the incompatible combinations of data type and encoder seen below:\n
                        {incompatibilities}
                    """))
                else:
                    encoding_dimension = "1D"
            else:
                encoding_dimension = "2D_singleColumn"
        else:
            encoding_dimension = "2D_multiColumn"
    return fitted_encoders, encoding_dimension


def transform_dynamicDimensions(
    fitted_encoders:list
    , encoding_dimension:str
    , samples_to_transform:object
):
    """
    After disabling LabelBinarizer and LabelEncoder from running on multiple 
    columns, every encoder seems to be fitting as "2D_multiColumn", but let's keep 
    the logic for other sklearn classes like feature extraction.
    """
    if (encoding_dimension == '2D_multiColumn'):
        # Our `to_arr` method fetches data as 2D. So it has 1+ columns. 
        encoded_samples = fitted_encoders[0].transform(samples_to_transform)
        encoded_samples = if_1d_make_2d(array=encoded_samples)
    elif (encoding_dimension == '2D_singleColumn'):
        # Means that `2D_multiColumn` arrays cannot be used as is.
        width = samples_to_transform.shape[1]
        if (width == 1):
            # It's already "2D_singleColumn"
            encoded_samples = fitted_encoders[0].transform(samples_to_transform)
            encoded_samples = if_1d_make_2d(array=encoded_samples)
        elif (width > 1):
            # Data must be fed into encoder as separate '2D_singleColumn' arrays.
            # Reshape "2D many columns" to “3D of 2D singleColumns” so we can loop on it.
            encoded_samples = samples_to_transform[None].T
            encoded_arrs = []
            for i, arr in enumerate(encoded_samples):
                encoded_arr = fitted_encoders[i].transform(arr)
                encoded_arr = if_1d_make_2d(array=encoded_arr)  
                encoded_arrs.append(encoded_arr)
            encoded_samples = np.array(encoded_arrs).T
            del encoded_arrs
    elif (encoding_dimension == '1D'):
        # From "2D_multiColumn" to "2D with 1D for each column"
        # This `.T` works for both single and multi column.
        encoded_samples = samples_to_transform.T
        # Since each column is 1D, we care about rows now.
        length = encoded_samples.shape[0]
        if (length == 1):
            # to get text feature_extraction working.
            encoded_samples = fitted_encoders[0].transform(encoded_samples[0])
            # Some of these 1D encoders also output 1D.
            # Need to put it back into 2D.
            encoded_samples = if_1d_make_2d(array=encoded_samples)  
        elif (length > 1):
            encoded_arrs = []
            for i, arr in enumerate(encoded_samples):
                encoded_arr = fitted_encoders[i].transform(arr)
                # Check if it is 1D before appending.
                encoded_arr = if_1d_make_2d(array=encoded_arr)              
                encoded_arrs.append(encoded_arr)
            # From "3D of 2D_singleColumn" to "2D_multiColumn"
            encoded_samples = np.array(encoded_arrs).T
            del encoded_arrs
    if (sparse.issparse(encoded_samples)):
        return encoded_samples.todense()
    return encoded_samples


def fit_labels(arr_labels:object, samples_train:list, labelcoder:object):
    """
    - `samples_train` are indices, not arrays.
    - All Label columns are always used during encoding.
    - Rows determine what fit happens.
    """
    preproc = labelcoder.sklearn_preprocess

    if (labelcoder.only_fit_train == True):
        labels_to_fit = arr_labels[samples_train]
    elif (labelcoder.only_fit_train == False):
        labels_to_fit = arr_labels

    fitted_encoders, encoding_dimension = fit_dynamicDimensions(
        sklearn_preprocess=preproc, samples_to_fit=labels_to_fit
    )
    return fitted_encoders, encoding_dimension


def fit_features(
    arr_features:object, samples_train:list,
    feature:object,
):
    featurecoders   = list(feature.featurecoders)
    fitted_encoders = []
    if (len(featurecoders) > 0):
        f_cols = feature.columns
        # For each featurecoder: fetch, transform, & concatenate matching features.
        # One nested list per Featurecoder. List of lists.
        for featurecoder in featurecoders:
            preproc = featurecoder.sklearn_preprocess

            if (featurecoder.only_fit_train == True):
                features_to_fit = arr_features[samples_train]
            elif (featurecoder.only_fit_train == False):
                features_to_fit = arr_features

            """
            Encoders want 2D data. 1-way trip; data destroyed after encoding.
            3D = seq | tab window
            4D = img | seq window
            5D = img window
            """			
            shape = features_to_fit.shape
            dim   = features_to_fit.ndim
            
            if (dim==3):
                rows            = shape[0] * shape[1]
                features_to_fit = features_to_fit.reshape(rows, shape[2])
            elif (dim==4):
                rows            = shape[0] * shape[1] * shape[2]
                features_to_fit = features_to_fit.reshape(rows, shape[3])
            elif (dim==5):
                rows            = shape[0] * shape[1] * shape[2] * shape[3]
                features_to_fit = features_to_fit.reshape(rows, shape[4])

            # Only fit these columns.
            matching_columns = featurecoder.matching_columns
            # Get the indices of the desired columns.
            col_indices = colIndices_from_colNames(
                column_names   = f_cols
                , desired_cols = matching_columns
            )

            # Data is coerced to 2D above
            features_to_fit = features_to_fit[:,col_indices]

            # Fit the encoder on the subset.
            fitted_coders, encoding_dimension = fit_dynamicDimensions(
                sklearn_preprocess = preproc
                , samples_to_fit   = features_to_fit
            )
            fitted_encoders.append(fitted_coders)
    return fitted_encoders, encoding_dimension


def transform_features(
    arr_features:object,
    fitted_encoders:list, feature:object 
):
    """
    Challenge: can't overwrite columns with data of different type (e.g. encoding object to int), 
    so they have to be pieced together.
    """
    featurecoders = list(feature.featurecoders)
    if (len(featurecoders) > 0):
        # Handle Sequence (part 1): reshape 3D to tall 2D for transformation.
        og_shape = arr_features.shape
        if (len(og_shape)==3):
            rows_2D = og_shape[0] * og_shape[1]
            arr_features = arr_features.reshape(rows_2D, og_shape[2])
        elif (len(og_shape)==4):
            rows_2D = og_shape[0] * og_shape[1] * og_shape[2]
            arr_features = arr_features.reshape(rows_2D, og_shape[3])

        f_cols               = feature.columns
        transformed_features = None #Used as a placeholder for `np.concatenate`.
        for featurecoder in featurecoders:
            idx                = featurecoder.idx
            fitted_coders      = fitted_encoders[idx]# returns list
            encoding_dimension = featurecoder.encoding_dimension
            
            # Only transform these columns.
            matching_columns = featurecoder.matching_columns
            # Get the indices of the desired columns.
            col_indices = colIndices_from_colNames(
                column_names   = f_cols
                , desired_cols = matching_columns
            )
            # Filter the array using those indices.
            features_to_transform = arr_features[:,col_indices]

            if (idx == 0):
                # It's the first encoder. Nothing to concat with, so just overwite the None value.
                transformed_features = transform_dynamicDimensions(
                    fitted_encoders        = fitted_coders
                    , encoding_dimension   = encoding_dimension
                    , samples_to_transform = features_to_transform
                )
            elif (idx > 0):
                encoded_features = transform_dynamicDimensions(
                    fitted_encoders        = fitted_coders
                    , encoding_dimension   = encoding_dimension
                    , samples_to_transform = features_to_transform
                )
                # Then concatenate w previously encoded features.
                transformed_features = np.concatenate(
                    (transformed_features, encoded_features)
                    , axis = 1
                )
        
        # After all featurecoders run, merge in leftover, unencoded columns.
        leftover_columns = featurecoders[-1].leftover_columns
        if (len(leftover_columns) > 0):
            # Get the indices of the desired columns.
            col_indices = colIndices_from_colNames(
                column_names   = f_cols
                , desired_cols = leftover_columns
            )
            # Filter the array using those indices.
            leftover_features = arr_features[:,col_indices]
                    
            transformed_features = np.concatenate(
                (transformed_features, leftover_features)
                , axis = 1
            )
        # Handle Sequence (part 2): reshape tall 2D back to original shape.
        if (len(og_shape)==3):
            transformed_features = arr_features.reshape(
                og_shape[0],
                og_shape[1],
                og_shape[2]
            )
        elif(len(og_shape)==4):
            transformed_features = arr_features.reshape(
                og_shape[0],
                og_shape[1],
                og_shape[2],
                og_shape[3]
            )
        elif(len(og_shape)==5):
            transformed_features = arr_features.reshape(
                og_shape[0],
                og_shape[1],
                og_shape[2],
                og_shape[3],
                og_shape[4],
            )
            
    elif (len(featurecoders) == 0):
        transformed_features = arr_features
    return transformed_features



"""
Plots
└── Documentation = https://aiqc.readthedocs.io/en/latest/notebooks/visualization.html
└── Data is prepared in the `plot_*` methods of the Queue and Predictor ORM classes.
"""
import plotly.graph_objects as go
import plotly.express as px
import plotly.figure_factory as ff
from re import sub


def set_dash_font(fig:object):
    """Dash knows about `/lab/assets` for Abel .tff, but regular charts don't"""
    return fig.update_layout(
        font         = dict(family="Abel, Avenir", size=14)
        , titlefont  = dict(family="Abel, Avenir")
        , hoverlabel = dict(font=dict(family="Abel, Avenir"))
    )


class Plot(object):
    def __init__(self):
        """
        - Defines a template that is passed to the other plots.
        - Font gets overriden to Abel below by Dash.
        """
        self.plot_template = dict(
            layout=go.Layout(
                font            = dict(family="Avenir", color='#FAFAFA', size=13)
                , title         = dict(x=0.05, y=1.0)
                , titlefont     = dict(family="Avenir")
                , title_pad     = dict(b=50, t=20)
                , plot_bgcolor  = '#182d41'
                , paper_bgcolor = '#182d41'
                , hovermode     = 'closest'
                , hoverlabel    = dict(
                    bgcolor = "#122536"
                    , font  = dict(family="Avenir", size=15)
                )
            )
        )


    def performance(
        self
        , dataframe:object
        , score_type:str
        , score_display:str
        , call_display:bool = True
        , height:int        = None
    ):
        fig = px.line(
            dataframe
            , title      = 'Model Metrics by Split'
            , x          = 'loss'
            , y          = score_type
            , color      = 'predictor_id'
            , height     = height
            , hover_data = ['predictor_id', 'split', 'loss', score_type]
            , line_shape = 'spline'
        )
        fig.update_traces(
            mode     = 'markers+lines'
            , line   = dict(width = 2)
            , marker = dict(
                size   = 8
                , line = dict(width=2, color='white')
            )
        )
        fig.update_layout(
            title_x       = 0.5#center
            , xaxis_title = "Loss"
            , yaxis_title = score_display
            , template    = self.plot_template
            , showlegend  = False#gets way too busy
        )

        fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
        fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
        
        if (call_display==True):
            fig.show()
        else:
            fig = set_dash_font(fig)
            return fig


    def learning_curve(
        self
        , dataframe:object
        , history_pairs:dict
        , skip_head:bool    = False
        , call_display:bool = True
    ):
        """Dataframe rows are epochs and columns are metric names."""
        if (skip_head==True):
            dataframe = dataframe.tail(round(dataframe.shape[0]*.85))
        # Spline seems to crash with too many points.
        if (dataframe.shape[0] >= 400):
            line_shape = 'linear'
        elif (dataframe.shape[0] < 400):
            line_shape = 'spline'

        # Create a plot for every pair.
        figs = []
        for train,val in history_pairs.items():
            metric_name = sub("_"," ",train[6:])
            df = dataframe[[train,val]]
            # The y axis is so variable that we need a range for it
            vals_all = list(df[train]) + list(df[val])
            tick_size = abs((max(vals_all) - min(vals_all))/7)
            fig = px.line(
                df, title=f"Training History: {metric_name}", line_shape=line_shape
            )
            fig.update_layout(
                title_y        = 1
                , xaxis_title  = "epochs"
                , yaxis_title  = metric_name
                , legend_title = None
                , height       = 400
                , template     = self.plot_template
                , yaxis        = dict(
                    side       = "right"
                    , tickmode = 'linear'
                    #when high loss, can't small fixed dtick. it freezes browser.
                    # `nticks` neither working here nor in `update_yaxes` below
                    , dtick    = tick_size
                )
                , legend = dict(
                    orientation = "h"
                    , yanchor   = "bottom"
                    , y         = 1.02
                    , xanchor   = "right"
                    , x         = 1
                )
            )
            fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
            fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
            if (call_display==True):
                fig.show()
            else:
                fig = set_dash_font(fig)
                figs.append(fig)
        if (call_display==False):
            return figs


    def confusion_matrix(
        self
        , cm_by_split
        , labels
        , call_display:bool = True
    ):
        figs = []
        for split, cm in cm_by_split.items():
            # change each element of z to type string for annotations
            cm_text = [[str(y) for y in x] for x in cm]

            fig = ff.create_annotated_heatmap(
                cm
                , x               = labels
                , y               = labels
                , annotation_text = cm_text
                , colorscale      = px.colors.sequential.BuGn
                , showscale       = False
                , colorbar        = dict(title='Count')
            )
            # add custom xaxis title
            fig.add_annotation(dict(
                font        = dict(color="white", size=15)
                , x         = 0.5
                , y         = 1.3
                , showarrow = False
                , text      = "Predicted Label"
                , xref      = "paper"
                , yref      = "paper"
            ))

            # add custom yaxis title
            fig.add_annotation(dict(
                font        = dict(color="white", size=15)
                , x         = -0.19
                , y         = 0.5
                # https://plotly.com/python/figure-structure/#positioning-with-paper-container-coordinates-or-axis-domain-coordinates
                , xref      = "x domain"
                , yref      = "paper"
                , showarrow = False
                , text      = "Actual Label"
                , textangle = -90

            ))

            fig.update_layout(
                title = dict(
                    text      = f"Confusion Matrix: {split.capitalize()}"
                    , y       = 0.0
                    , x       = 0.55
                    , xanchor = 'center'
                    , yanchor = 'bottom'
                )
                , template = self.plot_template
                , height   = 375  # if too small, it won't render in Jupyter.
                , yaxis    = dict(
                    tickmode     = 'linear'
                    , tick0      = 0.0
                    , dtick      = 1.0
                    , fixedrange = True#prevents zoom/pan
                    , tickfont   = dict(size=13)
                )
                , xaxis = dict(
                    categoryorder = 'category descending'
                    , fixedrange  = True#prevents zoom/pan
                    , tickfont    = dict(size=13)
                )
                #, margin = dict(r=75, l=150)#impacts y axis annotation x position
                , margin = dict(l=200)#impacts y axis annotation x position
            )

            fig.update_traces(
                hovertemplate = """predicted: %{x}<br>actual: %{y}<br>count: %{z}<extra></extra>"""
            )
            if (call_display==True):
                fig.show()
            else:
                fig = set_dash_font(fig)
                figs.append(fig)
        if (call_display==False):
            return figs


    def precision_recall(self, dataframe:object, call_display:bool=True):
        fig = px.line(
            dataframe
            , x     = 'recall'
            , y     = 'precision'
            , color = 'split'
            , title = 'Precision-Recall Curves'
        )
        fig.update_layout(
            legend_title = None
            , template   = self.plot_template
            , height     = 500
            , xaxis      = dict(title='Recall (Tₚ / Tₚ+Fₚ)')
            , yaxis      = dict(
                title      = 'Precision (Tₚ / Tₚ+Fₙ)'
                , side     = "right"
                , tickmode = 'linear'
                , tick0    = 0.0
                , dtick    = 0.05
            )
            , legend     = dict(
                orientation = "h"
                , yanchor   = "bottom"
                , y         = 1.02
                , xanchor   = "right"
                , x         = 1
            )
        )
        fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
        fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
        if (call_display==True):
            fig.show()
        else:
            fig = set_dash_font(fig)
            return fig


    def roc_curve(self, dataframe:object, call_display:bool=True):
        fig = px.line(
            dataframe
            , x     = 'fpr'
            , y     = 'tpr'
            , color = 'split'
            , title = 'Receiver Operating Characteristic (ROC) Curves'
        )
        fig.update_layout(
            template       = self.plot_template
            , legend_title = None
            , height       = 500
            , xaxis        = dict(
                title   = "False Positive Rate (FPR)"
                , tick0 = 0.00
                , range = [-0.025,1]
            )
            , yaxis      = dict(
                title      = "True Positive Rate (TPR)"
                , side     = "left"
                , tickmode = 'linear'
                , tick0    = 0.00
                , dtick    = 0.05
                , range    = [0,1.05]
            )
            , legend     = dict(
                orientation = "h"
                , yanchor   = "bottom"
                , y         = 1.02
                , xanchor   = "right"
                , x         = 1
            )
            , shapes=[
                dict(
                    type = 'line'
                    , y0=0, y1=1, x0=0, x1=1
                    , line = dict(dash='dot', width=2, color='#3b4043')
            )]
        )
        fig.update_xaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
        fig.update_yaxes(zeroline=False, gridcolor='#2c3c4a', tickfont=dict(color='#818487'))
        if (call_display==True):
            fig.show()
        else:
            fig = set_dash_font(fig)
            return fig
    

    def feature_importance(
        self, feature_impacts:object, feature_id:int,
        permute_count:int, top_n:int,
        height:int, margin_left:int,
        call_display:bool
    ):
        if (top_n is not None):
            title = f"Feature Importance <sub>(feature.id:{feature_id}, permute_count:{permute_count}, top_n:{top_n})</sub><br><br>"
        elif (top_n is None):
            title = f"Feature Importance <sub>(feature.id:{feature_id}, permute_count:{permute_count})</sub><br><br>"
        
        fig = go.Figure()
        for feature, impacts in feature_impacts.items():
            fig.add_trace(go.Box(x=impacts, name=feature))
        fig.update_layout(
            template     = self.plot_template
            , height     = height
            , title      = title
            , showlegend = False
            , margin     = dict(l=margin_left)
        )
        fig.update_xaxes(
            title = f"Importance<br><sup>[permuted column loss - training loss]</sup>",
            # ticks not really showing.
            tickangle=45, nticks=15, gridcolor='#2c3c4a'
        )
        if (call_display==True):
            fig.show()
        else:
            fig = set_dash_font(fig)
            return fig

"""
These plots don't use the template because they are formatted for a light UI
"""
def confidence_binary(
    sigmoid_curve:object
    , point:object
    , height:int
    , labels:list
    , call_display:bool = True
):
    fig = px.line(
        sigmoid_curve
        , title      = f"<b>Probability of {labels[1]}</b>: {point['Probability'][0]*100:.1f}%"
        , x          = 'x'
        , range_x    = [-6, 6]
        , y          = 'y'
        , range_y    = [0, 1]
        , line_shape = 'spline'
    ).update_layout(
        title_x      = 0.04
        , title_y    = 0.90
        , margin     = dict(l=0, r=0, t=0, b=0)
        , height     = height
        , xaxis      = dict(title=None, showticklabels=False, fixedrange=True)
        , yaxis      = dict(title=None, showticklabels=False, fixedrange=True)
        , hoverlabel = dict(font=dict(size=15))
        , title      = dict(font=dict(family='Avenir',size=15))
    ).update_traces(
        mode            = 'lines'
        , line          = dict(width=2, color='#004473')
        , hovertemplate = None
        , hoverinfo     = 'skip'
        
    # Shade & divide the quadrants
    ).add_hrect(
        y0=0.5
        , y1=1.0
        , line_width=0
        , fillcolor="yellow"
        , opacity=0.1 
    ).add_shape(
        type         = "rect"
        , x0         = 0
        , x1         = -6
        , y0         = 0.5
        , y1         = 1.0
        , fillcolor  = 'white'
        , line_color = 'white'
        , opacity    = 0.7
    ).add_shape(
        type         = "rect"
        , x0         = 0
        , x1         = 6
        , y0         = 0.5
        , y1         = 0
        , fillcolor  = 'white'
        , line_color = 'white'
        , opacity    = 0.6
    ).add_hline(
        y           = 0.5
        , line_dash = "dash"
        , fillcolor = 'gray'
        , opacity   = 0.7

    # Plot the confidence point
    ).add_traces(
        list(
            px.line(
                point
                , x = 'Logit(Probability)'
                , y = 'Probability',
            ).update_traces(
                mode     = 'markers'
                , marker = dict(size=12, color='#004473')
            ).select_traces()
        )
    
    # Label the quadrants
    ).add_annotation(
        text        = f"{labels[0]}"
        , xref      = "paper"
        , yref      = "paper"
        , x         = 0.2
        , y         = 0.23
        , showarrow = False
        , font      = dict(size=18, color='#09243d')
    ).add_annotation(
        text        = f"{labels[1]}"
        , xref      = "paper"
        , yref      = "paper"
        , x         = 0.7
        , y         = 0.77
        , showarrow = False
        , font      = dict(size=18, color='#05300c')
    )
    if (call_display==True):
        fig.show()
    else:
        fig = set_dash_font(fig)
        return fig


def confidence_multi(
    label_probabilities:object
    , height:int
    , call_display:bool = True
):
    fig = px.pie(
        label_probabilities
        , values                  = 'Probability'
        , names                   = 'Labels'
        , hole                    = .7
        , title                   = f"<b>Probabilities:</b>"
        , color_discrete_sequence = px.colors.qualitative.Pastel
    ).update_traces(
        textposition = 'outside'
    ).update_layout(
        legend_title_text = "Labels:"
        , height          = height
        , title_x         = 0.1
        , title_y         = 0.7
        , margin          = dict(l=0, r=0, t=15, b=15)
        , title           = dict(font=dict(size=15))
        , legend          = dict(
            yanchor   = "top"
            , y       = 1
            , xanchor = "right"
            , x       = 1.3
            , title   = dict(font = dict(size=15))
        )
    )
    if (call_display==True):
        fig.show()
    else:
        fig = set_dash_font(fig)
        return fig



from re import sub
from sklearn import metrics
import numpy as np


# Used by plots and UI.
metrics_classify = dict(
    accuracy  = 'Accuracy'
    , f1        = 'F1'
    , roc_auc   = 'ROC-AUC'
    , precision = 'Precision'
    , recall    = 'Recall'
)
metrics_classify_cols = list(metrics_classify.keys())
metrics_regress = dict(
    r2                   = 'R²'
    , mse                = 'MSE'
    , explained_variance = 'ExpVar'
)
metrics_regress_cols = list(metrics_regress.keys())
metrics_all          = {**metrics_classify, **metrics_regress}


def display_name(score_type:str):
    #`score_type` accesses df column, whereas `score_display` displays in plot
    score_display = sub("_", " ", score_type)
    if (score_display == "r2"):
        score_display = "R²"
    elif ((score_display=="roc auc") or (score_display=="mse")):
        score_display = score_display.upper()
    else:
        score_display = score_display.title()
    return score_display


def split_classification_metrics(labels_processed, predictions, probabilities, analysis_type):
    """
        - Be sure to register any new metrics in `metrics_classify` global.
        - Very rarely, these still fail (e.g. ROC when only 1 class of label is predicted).
    """
    if (analysis_type == "classification_binary"):
        average         = "binary"
        roc_average     = "micro"
        roc_multi_class = None
    elif (analysis_type == "classification_multi"):
        average         = "weighted"
        roc_average     = "weighted"
        roc_multi_class = "ovr"
        
    split_metrics = {}		
    # Let the classification_multi labels hit this metric in OHE format.
    split_metrics['roc_auc'] = metrics.roc_auc_score(labels_processed, probabilities, average=roc_average, multi_class=roc_multi_class)
    # Then convert the classification_multi labels ordinal format.
    if (analysis_type == "classification_multi"):
        labels_processed = np.argmax(labels_processed, axis=1)

    split_metrics['accuracy']  = metrics.accuracy_score(labels_processed, predictions)
    split_metrics['precision'] = metrics.precision_score(labels_processed, predictions, average=average, zero_division=0)
    split_metrics['recall']    = metrics.recall_score(labels_processed, predictions, average=average, zero_division=0)
    split_metrics['f1']        = metrics.f1_score(labels_processed, predictions, average=average, zero_division=0)
    return split_metrics


def split_regression_metrics(data, predictions):
    """Be sure to register any new metrics in `metrics_regress` global."""
    split_metrics = {}
    data_shape = data.shape
    # Unsupervised sequences and images have many data points for a single sample.
    # These metrics only work with 2D data, and all we are after is comparing each number to the real number.
    if (len(data_shape) == 5):
        data        = data.reshape(data_shape[0]*data_shape[1]*data_shape[2]*data_shape[3], data_shape[4])
        predictions = predictions.reshape(data_shape[0]*data_shape[1]*data_shape[2]*data_shape[3], data_shape[4])
    elif (len(data_shape) == 4):
        data        = data.reshape(data_shape[0]*data_shape[1]*data_shape[2], data_shape[3])
        predictions = predictions.reshape(data_shape[0]*data_shape[1]*data_shape[2], data_shape[3])
    elif (len(data_shape) == 3):
        data        = data.reshape(data_shape[0]*data_shape[1], data_shape[2])
        predictions = predictions.reshape(data_shape[0]*data_shape[1], data_shape[2])
    # These predictions are not persisted. Only used for metrics.
    split_metrics['r2']                 = metrics.r2_score(data, predictions)
    split_metrics['mse']                = metrics.mean_squared_error(data, predictions)
    split_metrics['explained_variance'] = metrics.explained_variance_score(data, predictions)
    return split_metrics


def split_classification_plots(labels_processed, predictions, probabilities, analysis_type):
    predictions     = predictions.flatten()
    probabilities   = probabilities.flatten()
    split_plot_data = {}
    
    if (analysis_type == "classification_binary"):
        labels_processed                    = labels_processed.flatten()
        split_plot_data['confusion_matrix'] = metrics.confusion_matrix(labels_processed, predictions)
        fpr, tpr, _                         = metrics.roc_curve(labels_processed, probabilities)
        precision, recall, _                = metrics.precision_recall_curve(labels_processed, probabilities)
    
    elif (analysis_type == "classification_multi"):
        # Flatten OHE labels for use with probabilities.
        labels_flat          = labels_processed.flatten()
        fpr, tpr, _          = metrics.roc_curve(labels_flat, probabilities)
        precision, recall, _ = metrics.precision_recall_curve(labels_flat, probabilities)

        # Then convert unflat OHE to ordinal format for use with predictions.
        labels_ordinal                      = np.argmax(labels_processed, axis=1)
        split_plot_data['confusion_matrix'] = metrics.confusion_matrix(labels_ordinal, predictions)

    split_plot_data['roc_curve']                           = {}
    split_plot_data['roc_curve']['fpr']                    = fpr
    split_plot_data['roc_curve']['tpr']                    = tpr
    split_plot_data['precision_recall_curve']              = {}
    split_plot_data['precision_recall_curve']['precision'] = precision
    split_plot_data['precision_recall_curve']['recall']    = recall
    return split_plot_data



"""PyTorch Regression with Tabular data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..utils.pytorch import fit
from ..orm import Dataset
# External modules
import torch
import torch.nn as nn
import torchmetrics as tm
from sklearn.preprocessing import PowerTransformer, MinMaxScaler, OrdinalEncoder


def fn_lose(**hp):
	if (hp['loss_type'] == 'mae'):
		# Same thng as mean absolute error.
		loser = nn.L1Loss()
	elif (hp['loss_type'] == 'mse'):
		loser = nn.MSELoss()
	return loser	


def fn_build(features_shape, label_shape, **hp):
	nc = hp['neuron_count']
	model = torch.nn.Sequential(
		nn.Linear(features_shape[-1], nc),
		nn.BatchNorm1d(nc,nc),
		nn.ReLU(),
		nn.Dropout(p=0.4),

		nn.Linear(nc, nc),
		nn.BatchNorm1d(nc,nc),
		nn.ReLU(),
		nn.Dropout(p=0.4),

		nn.Linear(nc, label_shape[-1])
	)
	return model


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	return fit(
		model, loser, optimizer, 
		train_features, train_label,
		eval_features, eval_label,
		epochs=10, batch_size=5,
		metrics=[tm.MeanSquaredError(),tm.R2Score(), tm.ExplainedVariance()]
	)


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=3):
	hyperparameters = {
		"neuron_count": [22]
		, "loss_type": ["mae","mse"]
	}
	
	file_path = datum.get_path('houses.csv')

	shared_dataset = Dataset.Tabular.from_path(
		file_path = file_path
	)

	pipeline = Pipeline(
		Input(
			dataset  = shared_dataset,
			encoders = [
				Input.Encoder(
					MinMaxScaler(copy=False)
					, include = False
					, dtypes = ['int64']
				),
				Input.Encoder(
					OrdinalEncoder()
					, include = False
					, dtypes = None
					, columns = None
				),
			]
		),

		Target(
			dataset = shared_dataset,
			column  = 'price',
			encoder = Target.Encoder(
				PowerTransformer(method='box-cox', copy=False)
			)
		),

		Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)
	)

	experiment = Experiment(
		Architecture(
			library           = "pytorch"
			, analysis_type   = "regression"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment
	


"""Functions that help with stratification, preprocessing, and inference."""
from .config import create_folder
from os import path
from textwrap import dedent
from torch import FloatTensor
from tqdm import tqdm
import numpy as np
import pandas as pd


default_interpolateKwargs = dict(
    method            = 'linear'
    , limit_direction = 'both'
    , limit_area      = None
    , axis            = 0
    , order           = 1
)


def listify(supposed_lst:object=None):
    """When only providing a single element, it's easy to forget to put it inside a list!"""
    if (supposed_lst is not None):
        if (not isinstance(supposed_lst, list)):
            supposed_lst = [supposed_lst]
        # If it was already a list, check it for emptiness and `None`.
        elif (isinstance(supposed_lst, list)):
            if (not supposed_lst):
                msg = "\nYikes - The list you provided is empty.\n"
                raise Exception(msg)
            if (None in supposed_lst):
                msg = f"\nYikes - The list you provided contained `None` as an element.\n{supposed_lst}\n"
                raise Exception(msg)
    # Allow entire list `is None` to pass through because we need it to trigger null conditions.
    return supposed_lst


def if_1d_make_2d(array:object):
    if (len(array.shape) == 1):
        array = array.reshape(array.shape[0], 1)
    return array


def size_shift_defined(size_window:int=None, size_shift:int=None):
    """Used by high level API classes."""
    if (
        ((size_window is None) and (size_shift is not None))
        or 
        ((size_window is not None) and (size_shift is None))
    ):
        msg = "\nYikes - `size_window` and `size_shift` must be used together or not at all.\n"
        raise Exception(msg)


def values_to_bins(array_to_bin:object, bin_count:int):
    """
    Overwites continuous Label values with bin numbers for statification & folding.
    Switched to `pd.qcut` because `np.digitize` never had enough samples in the up the leftmost/right bin.
    """
    # Make 1D for qcut.
    array_to_bin = array_to_bin.flatten()
    # For really unbalanced labels, I ran into errors where bin boundaries would be duplicates all the way down to 2 bins.
    # Setting `duplicates='drop'` to address this.
    bin_numbers  = pd.qcut(x=array_to_bin, q=bin_count, labels=False, duplicates='drop')
    # When the entire `array_to_bin` is the same, qcut returns all nans!
    if (np.isnan(bin_numbers).any()):
        bin_numbers = None
    else:
        # Convert 1D array back to 2D for the rest of the program.
        bin_numbers = np.reshape(bin_numbers, (-1, 1))
    return bin_numbers


def stratifier_by_dtype_binCount(
    stratify_dtype:object
    , stratify_arr:object
    , bin_count:int = None
):
    """Based on the dtype and bin_count determine how to stratify."""
    # Automatically bin floats.
    if np.issubdtype(stratify_dtype, np.floating):
        if (bin_count is None):
            bin_count = 3
        stratifier = values_to_bins(array_to_bin=stratify_arr, bin_count=bin_count)
    # Allow ints to pass either binned or unbinned.
    elif (
        (np.issubdtype(stratify_dtype, np.signedinteger))
        or
        (np.issubdtype(stratify_dtype, np.unsignedinteger))
    ):
        if (bin_count is not None):
            stratifier = values_to_bins(array_to_bin=stratify_arr, bin_count=bin_count)
        elif (bin_count is None):
            # Assumes the int is for classification.
            stratifier = stratify_arr
    # Reject binned objs.
    elif (np.issubdtype(stratify_dtype, np.number) == False):
        if (bin_count is not None):
            raise Exception(dedent("""
                Yikes - Your Label is not numeric (neither `np.floating`, `np.signedinteger`, `np.unsignedinteger`).
                Therefore, you cannot provide a value for `bin_count`.
            \n"""))
        elif (bin_count is None):
            stratifier = stratify_arr
    return stratifier, bin_count


def floats_only(label:object):
    # Prevent integer dtypes. It will ignore.
    label_dtypes = set(label.get_dtypes().values())
    for typ in label_dtypes:
        if (not np.issubdtype(typ, np.floating)):
            msg = f"\nYikes - Interpolate can only be ran on float dtypes. Your dtype: <{typ}>\n"
            raise Exception(msg)


def verify_interpolateKwargs(interpolate_kwargs:dict):
    if (interpolate_kwargs['method'] == 'polynomial'):
        msg = "\nYikes - `method=polynomial` is prevented due to bug: stackoverflow.com/questions/6722260.\n"
        raise Exception(msg)
    if ((interpolate_kwargs['axis'] != 0) and (interpolate_kwargs['axis'] != 'index')):
        # This makes it so that you can run on sparse indices.
        msg = "\nYikes - `axis` must be either 0 or 'index'.\n"
        raise Exception(msg)


def run_interpolate(dataframe:object, interpolate_kwargs:dict):
    # Interpolation does not require indices to be in order.
    dataframe = dataframe.interpolate(**interpolate_kwargs)
    return dataframe


def conditional_torch(arr:object, library:str=None):
    if (library=='pytorch'): 
        arr = FloatTensor(arr)
    return arr


def stage_data(splitset:object, fold:object):
    path_splitset = splitset.cache_path
    if (fold is not None):
        samples       = fold.samples
        idx           = fold.idx
        fold_idx      = f"fold_{idx}"
        path_fold     = path.join(path_splitset, fold_idx)
        create_folder(path_fold)
        fold_progress = f"📦 Caching Splits - Fold #{idx+1} 📦"
    else:
        samples       = splitset.samples
        path_fold     = path.join(path_splitset, "no_fold")
        create_folder(path_fold)
        fold_progress = "📦 Caching Splits 📦"
    key_train = splitset.key_train#fold-aware.
    """
    - Remember, you `.fit()` on either training data or the entire dataset (categoricals).
    - Then you transform the entire dataset because downstream processes may need the entire dataset:
    e.g. fit imputer to training data, then impute entire dataset so that categorical encoders can fit on entire dataset.
    - So we transform the entire dataset, then divide it into splits/ folds.
    - Then we convert the arrays to pytorch tensors if necessary. Subsetting with a list of indeces and `shape`
    work the same in both numpy and torch.
    """
    if (splitset.supervision == "supervised"):
        label      = splitset.label
        arr_labels = label.preprocess(
            samples     = samples
            , fold      = fold
            , key_train = key_train
        )

    # Expect different array shapes so use list, not array.
    features = []
    
    for feature in splitset.features:
        if (splitset.supervision == 'supervised'):
            arr_features = feature.preprocess(
                supervision = 'supervised'
                , fold      = fold
                , samples   = samples
                , key_train = key_train
            )
        elif (splitset.supervision == 'unsupervised'):
            # Remember, the unsupervised arr_labels may be different/shifted for forecasting.
            arr_features, arr_labels = feature.preprocess(
                supervision = 'unsupervised'
                , fold      = fold
                , samples   = samples
                , key_train = key_train
            )
        features.append(arr_features)
        # `arr_labels` is not appended because unsupervised analysis only supports 1 unsupervised feature.
    """
    - The samples object contains indices that we use to slice up the feature and label 
    arrays that are coming out of the preprocess() functions above
    - Keras multi-input models accept input as a list. Not using nested dict for multiple
    features because it would be hard to figure out feature.id-based keys on the fly.

    aiqc/cache/samples/splitset_uid
    └── fold_index | no_fold
        └── split
            └── label.npy
            └── feature_0.npy
            └── feature_1.npy
            └── feature_2.npy

    - 'no_fold' just keeps the folder depth uniform for regular splitsets
    - Tried label & feature folders, but it was too complex to fetch.
    """
    create_folder(path_splitset)

    for split, indices  in tqdm(
        samples.items(), desc=fold_progress, ncols=85
    ):	
        path_split = path.join(path_fold, split)
        create_folder(path_split)

        """
        `Object arrays cannot be saved when allow_pickle=True`
        "An object array is just a normal numpy array where the dtype is object"
        However, we expect all arrays to be numeric thanks to encoding.
        """
        path_label = path.join(path_split, "label.npy")
        np.save(path_label, arr_labels[indices], allow_pickle=True)

        for f, _ in enumerate(splitset.features):
            f_idx = f"feature_{f}.npy"
            path_feature = path.join(path_split, f_idx)
            np.save(path_feature, features[f][indices], allow_pickle=True)
    
    splitset.cache_hot = True
    splitset.save()


def fetchFeatures_ifAbsent(
    splitset:object
    , split:str 
    , train_features:object
    , eval_features:object
    , fold_id:int    = None
    , library:object = None
):
    """Check if this split is already in-memory. If not, fetch it."""
    key_trn  = splitset.key_train
    key_eval = splitset.key_evaluation
    
    fetch = True
    if (split==key_trn):
        if (train_features is not None):
            data  = train_features
            fetch = False
    elif (
        (split==key_eval) and (key_eval is not None)
        or
        ('infer' in split)
    ):
        if (eval_features is not None):
            data  = eval_features
            fetch = False
    
    if (fetch==True):
        data, _ = splitset.fetch_cache(
            fold_id          = fold_id
            , split          = split
            , label_features = 'features'
            , library        = library
        )
    return data


def fetchLabel_ifAbsent(
    splitset:object
    , split:str 
    , train_label:object
    , eval_label:object
    , fold_id:int    = None
    , library:object = None
):
    """Check if data is already in-memory. If not, fetch it."""
    key_trn  = splitset.key_train
    key_eval = splitset.key_evaluation
    
    fetch = True
    if (split==key_trn):
        if (train_label is not None):
            data  = train_label
            fetch = False
    elif (
        (split==key_eval) and (key_eval is not None)
        or
        ('infer' in split)
    ):
        if (eval_label is not None):
            data  = eval_label
            fetch = False
    
    if (fetch==True):
        data, _ = splitset.fetch_cache(
            fold_id          = fold_id
            , split          = split
            , label_features = 'label'
            , library        = library
        )
    return data


def columns_match(old:object, new:list):
    """
    - Set can be either Label or Feature.
    - Do not validate dtypes. If new columns have different NaN values, it changes their dtype.
      Encoders and other preprocessors ultimately record and use `matching_columns`, not dtypes.
    """
    cols_old = old.columns
    cols_new = new.columns
    if (cols_new != cols_old):
        msg = f"\nYikes - Columns do not match.\nNew: {cols_new}\n\nOld: {cols_old}.\n"
        raise Exception(msg)


def schemaNew_matches_schemaOld(splitset_new:object, splitset_old:object):
    # Get the new and old featuresets. Loop over them by index.
    features_new = splitset_new.features
    features_old = splitset_old.features

    if (len(features_new) != len(features_old)):
        msg = "\nYikes - Your new and old Splitsets do not contain the same number of Features.\n"
        raise Exception(msg)

    for i, f_new in enumerate(features_new):
        f_old = features_old[i]
        
        # --- Type & Dimensions ---
        typ_old = f_old.dataset.typ
        typ_new = f_new.dataset.typ
        if (typ_old != typ_new):
            msg = f"\nYikes - New Feature typ={typ_new} != old Feature typ={typ_old}.\n"
            raise Exception(msg)
        
        columns_match(f_old, f_new)
        
        if ((typ_new=='sequence') or (typ_new=='image')):
            rows_new = f_new.dataset.shape['rows']
            rows_old = f_old.dataset.shape['rows']
            if (rows_new != rows_old):
                msg = f"\nYikes - Row dimension does not match. New:{rows_new} vs Old:{rows_old}\n"
                raise Exception(msg)

        if (typ_new=='image'):
            channels_new = f_new.dataset.shape['channels']
            channels_old = f_old.dataset.shape['channels']
            if (channels_new != channels_old):
                msg = f"\nYikes - Image channel dimension does not match. New:{channels_new} vs Old:{channels_old}\n"
                raise Exception(msg)					
        
        # --- Window ---
        if (
            ((f_old.windows.count()>0) and (f_new.windows.count()==0))
            or
            ((f_new.windows.count()>0) and (f_old.windows.count()==0))
        ):
            msg = "\nYikes - Either both or neither of Splitsets can have Windows attached to their Features.\n"
            raise Exception(msg)

        if ((f_old.windows.count()>0) and (f_new.windows.count()>0)):
            window_old = f_old.windows[-1]
            window_new = f_new.windows[-1]
            if (
                (window_old.size_window != window_new.size_window)
                or
                (window_old.size_shift != window_new.size_shift)
            ):
                msg = "\nYikes - New Window and old Window schemas do not match.\n"
                raise Exception(msg)

    """
    - Only verify Labels if the inference new Splitset provides Labels.
    - Otherwise, it may be conducting pure inference.
    - Labels can only be 'tabular' so don't need to validate type
    """
    label = splitset_new.label
    if (label is not None):
        if (splitset_old.supervision == 'unsupervised'):
            msg = "\nYikes - New Splitset has Labels, but old Splitset does not have Labels.\n"
            raise Exception(msg)
        elif (splitset_old.supervision == 'supervised'):
            label_Old =  splitset_old.label

        columns_match(label_Old, label)



"""TensorFlow Multi-label Classification with Tabular data"""
# Internal modules
from ..mlops import Pipeline, Input, Target, Stratifier, Experiment, Architecture, Trainer
from .. import datum
from ..orm import Dataset
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import StandardScaler, RobustScaler, OneHotEncoder


def fn_build(features_shape, label_shape, **hp):
	m = tf.keras.models.Sequential()
	m.add(l.Input(shape=features_shape))
	m.add(l.Dense(units=hp['neuron_count'], activation='relu', kernel_initializer='he_uniform'))
	m.add(l.Dropout(0.2))
	m.add(l.Dense(units=label_shape[-1], activation='softmax'))
	return m


def fn_optimize(**hp):
	optimizer = tf.keras.optimizers.Adamax(hp['learning_rate'])
	return optimizer


def fn_train(
	model, loser, optimizer,
	train_features, train_label,
	eval_features, eval_label,
	**hp
):
	model.compile(
		loss = loser
		, optimizer = optimizer
		, metrics = ['accuracy']
	)

	model.fit(
		train_features, train_label
		, validation_data = (eval_features, eval_label)
		, verbose = 0
		, batch_size = hp['batch_size']
		, epochs = hp['epoch_count']
		, callbacks=[tf.keras.callbacks.History()]
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=2):
	hyperparameters = {
		"neuron_count": [9]
		, "batch_size": [3]
		, "learning_rate": [0.05]
		, "epoch_count": [10]
	}

	# Note: iris 10x has ordinal labels, not text.
	if (fold_count is not None):
		file_path = datum.get_path('iris_10x.tsv')
		name = "iris"
		description = "Expanded sample population for cross validation"
	else:
		file_path = datum.get_path('iris.tsv')
		name = "iris"
		description = "Just large enough to be representative of population"
	
	shared_dataset = Dataset.Tabular.from_path(
		file_path = file_path
		, name = name
		, description = description
	)

	pipeline = Pipeline(
		Input(
			dataset  = shared_dataset,
			encoders = [
				Input.Encoder(
					StandardScaler(copy=False),
					columns = ['petal_width']
				),
				Input.Encoder(
					RobustScaler(copy=False),
					dtypes = ['float64']
				)
			]
		),

		Target(
			dataset = shared_dataset,
			column  = 'species',
			encoder = Target.Encoder(OneHotEncoder(sparse=False))
		),

		Stratifier(
			size_test       = 0.11, 
			size_validation = 0.21,
			fold_count      = fold_count
		)    
	)
	experiment = Experiment(
		Architecture(
			library           = "keras"
			, analysis_type   = "classification_multi"
			, fn_build        = fn_build
			, fn_train        = fn_train
			, hyperparameters = hyperparameters
		),
		
		Trainer(
			pipeline       = pipeline
			, repeat_count    = repeat_count
			, permute_count   = permute_count
			, search_percent  = None
		)
	)
	return experiment
